from Bio.PDB import *
import pandas as pd
import numpy as np
import urllib.request
import os.path
import os
import re

def percent(genotype, phenotype,snps2aa):
	geno  = genotype.values
	geno_sum = geno.sum(axis=0)
	pheno = phenotype.values.reshape((1,-1))

	percent = pheno.dot(geno).reshape(-1)
	percent = percent / geno_sum

	percent_df = pd.DataFrame(index=genotype.columns)
	percent_df.loc[:,'es'] = percent
	snp_df = pd.merge(percent_df, snps2aa,left_index=True, right_on='varcode')
	
	return snp_df

def parser_vep(vep_input):
	
	with open(vep_input)as in_file:
		for line in in_file:
			if re.match("^##INFO=<ID=CSQ",line): 
				csq_header_raw = line.split('\"')[1].split(':')[1]
				csq_header = csq_header_raw.split('|')
			if re.match("^#CHROM",line): 
				vcf_header_raw = line.strip('\n').split('\t')
				vcf_header = vcf_header_raw[0:8]
	vep_df = pd.read_csv(vep_input,comment='#',sep='\t',header=None,names=vcf_header, \
				usecols=range(8))
	temp_df = vep_df['INFO'].str.split(',', expand=True).add_prefix('csq')
	vep_df = vep_df.join(temp_df)
	temp_header = temp_df.columns

	csq_df = vep_df.melt(id_vars=vcf_header, value_vars=temp_header,value_name='csq')
	temp_df = csq_df['csq'].str.split('|',expand=True)
	if temp_df.shape[1] < len(csq_header): return pd.DataFrame()
	temp_df.columns = csq_header
	
	csq_df[csq_header] = temp_df[csq_header]
	csq_df_filter = csq_df[ (csq_df['Consequence']=='missense_variant') & \
				(csq_df['CANONICAL']=='YES')]
	pdbentry = csq_df_filter['SWISSPROT'].str.split('_',1, expand=True)
	csq_df_filter.loc[:,'SWISSPROT'] = pdbentry[0]
	csq_df_filter.loc[:,'ID']=csq_df_filter[['#CHROM','POS','REF','ALT']].astype(str).agg(':'.join,axis=1)

	#extend the csq dataframe to all possible amino acid change
	#ref as destinate allele
	csq_df_filter_part1 = csq_df_filter.copy()
	csq_df_filter_part1['des_allele'] = csq_df_filter_part1['REF']
	csq_df_filter_part1['Amino_acids'] = csq_df_filter['Amino_acids'].str[::-1]

	#alt as destinate allele
	csq_df_filter_part2 = csq_df_filter.copy()
	csq_df_filter_part2['des_allele'] = csq_df_filter_part2['ALT']
	csq_df_filter_part2['Amino_acids'] = csq_df_filter['Amino_acids']
	csq_df_filter_ext = pd.concat([csq_df_filter_part1,csq_df_filter_part2],ignore_index=True)

	csq_df_filter_ext['full_id'] = csq_df_filter_ext[['ID','des_allele']].astype(str).agg('_'.join,axis=1)
	return csq_df_filter_ext[['full_id','Feature','SWISSPROT','Protein_position','Amino_acids']]


def filter_snps2aa(snps2aa_noidx, pdb='None'):
	# find the protein with most varaints mapped
	if not pdb:
		uniq_map = snps2aa_noidx.groupby(['structure','chain'])['varcode'].count().reset_index()
		iline = uniq_map['varcode'].argmax()
		pdb = uniq_map.loc[iline,'structure']

	snps2aa = snps2aa_noidx.loc[snps2aa_noidx['structure']==pdb]

	snps_mapped = snps2aa['varcode'].unique().tolist()

	idx_tab = pd.DataFrame()
	idx_tab.loc[:,'id'] = list(range(len(snps_mapped)))
	idx_tab.loc[:,'varcode'] = snps_mapped
	snps2aa = pd.merge(snps2aa, idx_tab, on='varcode')

	return snps2aa,pdb

def snps_to_aa(vep,genotype,map_to_pdb,database='pdb'): 

	snps = set(vep['full_id'].values).intersection(genotype.columns.tolist())
	vep_mapping_processing = vep.loc[vep['full_id'].isin(snps)]
	vep_mapping_processing.loc[:,'SWISSPROT'] = vep_mapping_processing['SWISSPROT'].str.split('.').str[0]
	
	if database == "pdb":
		vep_mapping = pd.merge(vep_mapping_processing, map_to_pdb, on='SWISSPROT')
		out_df = map_PDB_structure(vep_mapping)	

	elif database == "alphafold":
		vep_mapping = vep_mapping_processing.copy()
		vep_mapping.loc[:,'structure'] = vep_mapping['SWISSPROT'] 
		out_df = map_alphafold_structure(vep_mapping)

	return out_df
			
def map_PDB_structure(vep_mapping):

	pdb_struct_dir='ref/pdb'
	if not os.path.exists(pdb_struct_dir):os.makedirs(pdb_struct_dir)	

	ori_cols = ['full_id','structure','chain','Protein_position','x','y','z','Amino_acids']
	new_cols = ['varcode','structure','chain','structure_position','x','y','z','aa']

	if len(vep_mapping['structure']) == 0: 
		print("no PDB structure mapped to snps")
		return pd.DataFrame(columns=new_cols)

	vep_mapping.loc[:,'x'] = np.nan
	vep_mapping.loc[:,'y'] = np.nan
	vep_mapping.loc[:,'z'] = np.nan
	
	parser = PDBParser()
	structure_list = {}
	for pdb in set(vep_mapping['structure'].values):
		try:
			pdbl=PDBList()
			pdbl.retrieve_pdb_file(pdb, pdir='ref/pdb', file_format='pdb',obsolete=False)
			structure = parser.get_structure(pdb, "ref/pdb/pdb%s.ent"%pdb)
			structure_list[pdb] = structure
		except: continue

	for irow, row in vep_mapping.iterrows():	
		entry = row['structure']
		chain = row['chain']
		residue = int(row['Protein_position'])
		try:
			structure = structure_list.get(entry)
			# ensure the expression source
			org_sys = structure.header['source']['1']['organism_scientific'].replace(' ','').split(',')
			if len(org_sys) > 1 or org_sys[0] != 'homosapiens': continue 

			atom  = structure[0][chain][residue]["CA"]
			coord = atom.get_coord()
			vep_mapping.loc[irow,['x','y','z']] = coord
		except: continue
	out_df = vep_mapping[ori_cols]
	out_df.columns = new_cols

	return	out_df.dropna().drop_duplicates().reset_index(drop=True)

def map_alphafold_structure(vep_mapping):

	af_struct_dir='ref/alphafold'
	if not os.path.exists(af_struct_dir):os.makedirs(af_struct_dir)	

	ori_cols = ['full_id','structure','chain','Protein_position','x','y','z','Amino_acids']
	new_cols = ['varcode','structure','chain','structure_position','x','y','z','aa']

	if len(vep_mapping['structure']) == 0: 
		print("no alphafold structure mapped to snps")
		return pd.DataFrame(columns=new_cols)

	vep_mapping.loc[:,'x'] = np.nan
	vep_mapping.loc[:,'y'] = np.nan
	vep_mapping.loc[:,'z'] = np.nan

	parser = PDBParser()
	structure_list = {}
	for uniprot in vep_mapping['structure'].unique():
		try:
			chain_id = []
			url='https://alphafold.ebi.ac.uk/files/AF-%s-F1-model_v1.pdb'%(uniprot)
			urllib.request.urlretrieve(url, "ref/alphafold/%s.pdb"%uniprot)
			structure = parser.get_structure(uniprot, "ref/alphafold/%s.pdb"%uniprot)
			structure_list[uniprot] = structure
			for chain in structure[0].get_list():
				chain_id += chain.get_id()
			vep_mapping['chain_list'] = ','.join(chain_id)
		except: continue
	vep_mapping.loc[:,'chain'] = vep_mapping.chain_list.str.split(',')
	vep_mapping = vep_mapping.explode('chain')

	for irow, row in vep_mapping.iterrows():	
		entry = row['structure']
		chain = row['chain']
		residue = int(row['Protein_position'])
		try:
			structure = structure_list.get(entry)
			resol = structure.header["resolution"]
			atom  = structure[0][chain][residue]["CA"]
			coord = atom.get_coord()
			vep_mapping.loc[irow,['x','y','z']] = coord
		except: continue
	out_df = vep_mapping[ori_cols]
	out_df.columns = new_cols

	return	out_df.dropna().drop_duplicates().reset_index(drop=True)

def parser_vcf(genotype_file,phenotype_file,cov_file,cov_list,freq):

	# process genotype file
	genotype_raw=pd.read_csv(genotype_file,sep='\s+|\t|,',engine='python',index_col=1)
	genotype = genotype_raw.iloc[:,5:]
	genotype.index = genotype.index.map(str)

	# process covariates
	if cov_file:
		cov_raw = pd.read_csv(cov_file,sep='\s+|\t|,',engine='python',index_col=0)
		cov_raw.dropna(inplace=True)
		cov = cov_raw.loc[:,cov_list]
		cov.index   = cov.index.map(str)
	else: cov = None

	# process phenotype files
	phenotype = pd.read_csv(phenotype_file,sep='\s+|\t|,',engine='python',index_col=0)
	phenotype.dropna(inplace=True)
	phenotype.index = phenotype.index.map(str)

	if phenotype.shape[1] < 2: phenotype = phenotype.transpose()
	phenotype_ind = phenotype.columns.tolist()

	genotype_ind = genotype.index.tolist()
	individual = set(genotype_ind).intersection(phenotype_ind)
	if cov_file:	
		cov_ind = cov.index.tolist()
		individual = set(individual).intersection(cov_ind)
		cov = cov.loc[individual]

	## calculate freq
	freqs = genotype.sum(axis=0,skipna=True) 
	freqs = freqs /(2 * genotype.count())
	
	freqs_df = freqs.to_frame(name='freq')
	# secure a minor allele count matrix
	
	freqs_df['index'] = freqs_df.index		
	freqs_df[['ID','c_allele']] = freqs_df['index'].str.split('_',expand=True)
	freqs_df[['#CHROM','POS','REF','ALT']] = freqs_df['ID'].str.split(':',expand=True)

	genotype_processing = genotype.T
	flip_allele_bool = freqs > (1-freq)
	genotype_processing.loc[flip_allele_bool] = genotype_processing.loc[flip_allele_bool].replace({2:0, 0:2})
	freqs_df.loc[flip_allele_bool,'index'] = freqs_df.loc[flip_allele_bool,['ID','ALT']].astype(str).agg('_'.join,axis=1)

	genotype_processing.set_index(freqs_df['index'].values, inplace=True)
	ma_genotype = genotype_processing.T

	ma_freqs = ma_genotype.sum(axis=0,skipna=True) 
	ma_freqs = ma_freqs /(2 * ma_genotype.count())
		
	ma_freqs_subset = ma_freqs[(ma_freqs < freq) & (ma_freqs > 0)]
	snps = ma_freqs_subset.index.tolist()

	ma_genotype.fillna(0, inplace=True)
	return ma_genotype.loc[individual,snps],ma_freqs[snps],phenotype[individual], cov

if __name__ == "__main__":
	main()





import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from sklearn.cluster import DBSCAN
import pymol
from pymol import cmd
import numpy as np 
import pandas as pd
import pickle
import sys

class cluster:

	def __init__(self,genotype,snps2aa,phenotype,distance,pdb):
		self.genotype  = genotype
		self.phenotype = phenotype
		self.snps2aa   = snps2aa
		self.distance  = distance
		self.pdb       = pdb

		self.snps2aa = self.snps2aa[self.snps2aa['structure']==self.pdb]
		snps_sum = self.genotype.sum(axis=0)
		snps_sum = snps_sum[snps_sum>0]
		snps2aa_subset = self.snps2aa.merge(snps_sum.to_frame(),left_on='varcode',right_index=True)
		print('snp:', len(snps_sum))
		print('mapped snp:',snps2aa_subset.shape[0])
		print('#sub:', self.phenotype.shape[1])
		print('#case sub:', np.sum(self.phenotype.values))

	def cluster(self,matrix):
		clustering = DBSCAN(metric='precomputed',eps=14,min_samples=3)
		clustering.fit(matrix)
		label = clustering.labels_
		return label

	def cluster_analysis(self):
		case_id, ctrl_id, snp_df_sub = self.score_on_var()	
		case_mtx = self.distance[case_id,:][:,case_id]
		ctrl_mtx = self.distance[ctrl_id,:][:,ctrl_id]
		# case variant
		case_label = self.cluster(case_mtx)
		# ctrl variant
		ctrl_label = self.cluster(ctrl_mtx)

		case_cluster = pd.DataFrame(data=case_label,index=case_id, columns=['grp'])
		case_cluster_df = pd.merge(case_cluster,self.snps2aa,left_index=True,right_on='id')
		ctrl_cluster = pd.DataFrame(data=ctrl_label,index=ctrl_id, columns=['grp'])
		ctrl_cluster_df = pd.merge(ctrl_cluster,self.snps2aa,left_index=True,right_on='id')

		case_mtx_df = pd.merge(case_cluster_df, snp_df_sub, on='id')
		ctrl_mtx_df = pd.merge(ctrl_cluster_df, snp_df_sub, on='id')

		case_mtx_df.to_csv('%s_case.csv'%self.pdb, index=False)
		ctrl_mtx_df.to_csv('%s_ctrl.csv'%self.pdb,index=False)

		return case_cluster_df, ctrl_cluster_df
	
	def plot_cluster(self,outfile):
		case_cluster_df, ctrl_cluster_df = self.cluster_analysis()
		print(case_cluster_df)
		case_cluster_df.to_csv('%s.csv'%outfile,mode='w')
		print(ctrl_cluster_df)
		ctrl_cluster_df.to_csv('%s.csv'%outfile,mode='a')
		# for case
		n_case_grp = case_cluster_df['grp'].max() + 1
		case_map = plt.get_cmap("OrRd_r", n_case_grp + 3)
		case_color_list = [case_map(i)[:3] for i in range(n_case_grp)]
		case_color_dict = dict(zip(list(range(n_case_grp)),case_color_list))

		#pymol.finish_launching()
		cmd.reinitialize()
		try:
			cmd.load('ref/pdb/pdb%s.ent'%self.pdb, self.pdb)
		except:
			cmd.load('ref/alphafold/%s.pdb'%self.pdb, self.pdb)
		
		cmd.show_as('cartoon',self.pdb)
		cmd.color('white',self.pdb)
		
		case_cluster_df_grp = case_cluster_df.groupby('grp')
		for igrp, case_grp in case_cluster_df_grp:
			if igrp == -1: continue
			color_name = 'case'+str(igrp)
			cmd.set_color(color_name, case_color_dict[igrp]) 
			for i, row in case_grp.iterrows():
				resi  = row['structure_position']
				chain = row['chain']
				selec = 'snp%s_case%s'%(i,igrp)
				selec_atom = 'snp_atom%s_case%s'%(i,igrp)
				cmd.select(selec,'name ca and resi %s and chain %s'%(resi,chain))
				cmd.create(selec_atom, selec)
				cmd.set("sphere_scale",0.8)
				cmd.show('sphere',selec_atom)
				cmd.color(color_name, selec_atom)

		# for ctrl
		n_ctrl_grp = ctrl_cluster_df['grp'].max() + 1
		ctrl_map = plt.get_cmap("PuBu_r", n_ctrl_grp + 3)
		ctrl_color_list = [ctrl_map(i)[:3] for i in range(n_ctrl_grp)]
		ctrl_color_dict = dict(zip(list(range(n_ctrl_grp)),ctrl_color_list))

		ctrl_cluster_df_grp = ctrl_cluster_df.groupby('grp')
		for igrp, ctrl_grp in ctrl_cluster_df_grp:
			if igrp == -1: continue
			color_name = 'ctrl'+str(igrp)
			cmd.set_color(color_name, ctrl_color_dict[igrp]) 
			for i, row in ctrl_grp.iterrows():
				resi  = row['structure_position']
				chain = row['chain']
				selec = 'snp%s_ctrl%s'%(i,igrp)
				selec_atom = 'snp_atom%s_ctrl%s'%(i,igrp)
				cmd.select(selec,'name ca and resi %s and chain %s'%(resi,chain))
				cmd.create(selec_atom, selec)
				cmd.set("sphere_scale",0.6)
				cmd.show('sphere',selec_atom)
				cmd.color(color_name,selec_atom)

		cmd.bg_color("white")
		cmd.zoom()
		cmd.orient()
		cmd.png('%s_cluster.png'%outfile, width=2400, height=2400, dpi=300, ray=1)
		cmd.save('%s_cluster.pse'%outfile)

	def plot(self,outfile):
		ctrl_id, case_id, snp_df_sub = self.score_on_var()	
		df = pd.merge(snp_df_sub, self.snps2aa,on='id')

		print(self.pdb)
	
		#pymol.finish_launching()
		cmd.reinitialize()
		try:
			cmd.load('ref/pdb/pdb%s.ent'%self.pdb, self.pdb)
		except:
			cmd.load('ref/alphafold/%s.pdb'%self.pdb, self.pdb)
		
		cmd.show_as('cartoon',self.pdb)
		cmd.color('white',self.pdb)

		cmap = mcolors.LinearSegmentedColormap.from_list("", ["tab:blue","white","tab:red"])
		norm = mcolors.Normalize(vmin=0, vmax=1)

		for i, row in df.iterrows():
			resi  = row['structure_position']
			chain = row['chain']
			pheno = row['es']
			selec = 'snp%s'%i
			selec_atom = 'snp_atom%s'%i

			icolor = cmap(norm(pheno))[0:3]
			color_name = 'color%s'%i
			cmd.set_color(color_name, icolor)

			cmd.select(selec,'name ca and resi %s and chain %s'%(resi,chain))
			cmd.create(selec_atom, selec)
			cmd.set("sphere_scale", 0.6)
			cmd.show('sphere', selec_atom)
			cmd.color(color_name, selec_atom)

		cmd.bg_color("white")
		cmd.zoom()
		cmd.orient()
		cmd.save('%s.pse'%outfile)
		cmd.png('%s.png'%outfile, width=2400, height=2400, dpi=300, ray=1)

	def score_on_var(self):
		geno  = self.genotype.values
		geno_sum = geno.sum(axis=0)
		pheno = self.phenotype.values.reshape((1,-1))

		percent = pheno.dot(geno).reshape(-1)
		percent = percent / geno_sum

		percent_df = pd.DataFrame(index=self.genotype.columns)
		percent_df['es'] = percent
	
		snp_df = pd.merge(percent_df, self.snps2aa,left_index=True, right_on='varcode')
		snp_df_sub = snp_df.loc[snp_df['structure']==self.pdb,['id','es']]
		snp_df_sub.dropna(inplace=True)

		ctrl_id = snp_df_sub.loc[snp_df_sub['es']<=0.5, 'id']
		case_id = snp_df_sub.loc[snp_df_sub['es']>0.5, 'id']
		return case_id, ctrl_id, snp_df_sub

def main():

	infile  = sys.argv[1]
	pdb = sys.argv[2]
	prefix = infile.split('.')[0]

	with open(infile, 'rb') as f:
		genotype = pickle.load(f)
		phenotype = pickle.load(f)
		snps2aa = pickle.load(f)
		distance_mat = pickle.load(f)

	outfig = prefix
	cls1 = cluster(genotype,snps2aa,phenotype,distance_mat,pdb)
	cls1.plot(outfig)
	cls1.plot_cluster(outfig)	

# main body
if __name__ == "__main__":
	main()

